{% load static %}
{% load i18n %}

<div class="form-group d-flex">
    <button id="updatebtn" type="submit" class="btn btn-default" data-toggle="tooltip" data-placement="bottom" title="Next"><img src="{% static 'gui2/appstore2/images/play.svg' %}" alt="" /> {% translate 'Next' %}</button>
</div>
<div class="form-group d-flex">
{#    <button name="submit" id="updatebtn" value="select" type="submit" class="btn btn-default" data-toggle="tooltip" data-placement="bottom" title="Submit"><img src={% static 'gui2/appstore2/images/'|add:icon|add:'.svg' %}</img>Select</button>#}
</div>